#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include "macres.h"
#include "cicn.h"
#include "gtk_hlist.h"

static char *file = NULL;
static macres_file *cicns = NULL;

static GdkGC *mask_gc;

static GtkWidget *icon_list;
static GtkWidget *icon_label;
static GtkWidget *window;

#define TYPE_cicn	0x6369636e

static void list_icons (void);

static int listsorthelper (GtkHList *hlist,
				GtkHListRow *ptr1,
				GtkHListRow *ptr2)
{
	int i1=atoi(GTK_HELL_TEXT(ptr1->cell[1])->text);
	int i2=atoi(GTK_HELL_TEXT(ptr2->cell[1])->text);
	if (i1 < i2) return -1;
	if (i1 > i2) return 1;
	return 0;
}

static void create_ifn(char *str) 
{
	int fd;

	if(file) {
		g_free(file);
		file = NULL;
	}
	if(cicns) {
		macres_file_delete(cicns);
		cicns = NULL;
	}

	file = strdup(str);
	fd = open(file, O_RDONLY);
	if(fd < 0) {
		cicns = NULL;
		g_free(file);
		file = NULL;
	}
	else {
		cicns = macres_file_open(fd);
		close(fd);
	}
}

static void open_icons(GtkWidget *btn, GtkWidget *filesel)
{
	char *path = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
	
	create_ifn(path);
	gtk_hlist_clear(GTK_HLIST(icon_list));

	list_icons();

	gtk_widget_destroy(filesel);
}

static void open_iconfile(guint callback_data, guint callback_action, 
						  GtkWidget *widget)
{
	GtkWidget *filesel = gtk_file_selection_new("Open...");
	
	gtk_signal_connect_object(GTK_OBJECT(filesel), "destroy", 
							  (GtkSignalFunc)gtk_widget_destroy, 
							  GTK_OBJECT(filesel));
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(
											 filesel)->cancel_button),
							  "clicked", (GtkSignalFunc)gtk_widget_destroy,
							  GTK_OBJECT(filesel));

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button), 
					   "clicked", GTK_SIGNAL_FUNC(open_icons), filesel);

	gtk_widget_show_all(filesel);
	gtk_grab_add(filesel);
}

static int about_open = 0;

static void close_about(GtkWidget *widget, GtkWidget *dialog)
{
	about_open = 0;
	gtk_widget_destroy(dialog);
}

static void create_about_window(guint callback_data, guint callback_action, 
								GtkWidget *widget)
{
	const char *authors[] = {
		"Misha Nasledov <misha@nasledov.com>",
		"Aaron Lehmann <aaronl@vitelus.com>",
		"Ryan Nielsen <ran@krazynet.com>", 0 };
	
	GtkWidget *about_box;
	GtkWidget *close;
	GtkWidget *vbox;
	GtkWidget *lbl;
	GtkWidget *hbox;
	GtkWidget *avbox;
	char *lbl_str;
	int i;
	
	if(about_open) return;
	
	about_open = 1;

	about_box = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(about_box), "About GCicnV");

	close = gtk_button_new_with_label("OK");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(about_box)->action_area), close, 0, 
					   0, 5);

	gtk_signal_connect(GTK_OBJECT(close), "clicked", 
					   GTK_SIGNAL_FUNC(close_about), about_box);
	gtk_signal_connect(GTK_OBJECT(about_box), "delete_event", 
					   GTK_SIGNAL_FUNC(close_about), about_box);

	vbox = GTK_DIALOG(about_box)->vbox;
	gtk_box_set_spacing(GTK_BOX(vbox), 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
	
	lbl_str = g_strdup_printf("%s %s - A Gtk CICN Viewer", PACKAGE, VERSION);
	lbl = gtk_label_new(lbl_str);

	hbox  = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	lbl = gtk_label_new("Copyright (C) 2001 Misha Nasledov");
	hbox = gtk_hbox_new(0, 0);
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	
	hbox = gtk_hbox_new(0, 0);
	avbox = gtk_vbox_new(0, 5);
	gtk_box_pack_start(GTK_BOX(hbox), avbox, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 0, 0, 0);

	hbox = gtk_hbox_new(0, 0);
	lbl = gtk_label_new("Authors: ");
	gtk_box_pack_start(GTK_BOX(hbox), lbl, 0, 0, 0);
	gtk_box_pack_start(GTK_BOX(avbox), hbox, 0, 0, 0);
	
	for(i = 0; authors[i]; i++) {
		lbl = gtk_label_new(authors[i]);
		gtk_box_pack_start(GTK_BOX(hbox), lbl, 1, 1, 0);

		hbox = gtk_hbox_new(0, 0);
		gtk_box_pack_start(GTK_BOX(avbox), hbox, 0, 0, 0);
	}
		
	gtk_widget_show_all(about_box);
}

static void quit_gcicnv(guint callback_data, guint callback_action, 
						GtkWidget *widget)
{
	gtk_main_quit();
}

static GtkItemFactoryEntry menu_items[] = {
	{"/_File", 0, 0, 0, "<Branch>" },
	{"/File/_Open", "<control>O", open_iconfile, 0},
	{"/File/_Quit", "<control>Q", quit_gcicnv, 0},
	{"/_Help", 0, 0, 0, "<LastBranch>"},
	{"/Help/_About", 0, create_about_window, 0}
};

static int nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

static void list_icons (void)
{
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	gint row;
	gchar *text[2] = {0, 0};
	guint32 icon;
	unsigned int nfound = 0;
	char buf[16];
	guint16 nres = 0;
	GdkColormap *colormap = gtk_widget_get_colormap(icon_list);
	GdkVisual *visual = gtk_widget_get_visual(icon_list);
	GdkGC *gc = NULL;
	GdkImage *image;
	GdkImage *maskim;
	int off;
	char *title;

	if(!cicns)
		return;

	if(file) {
		title = g_strdup_printf("GCicnV - %s", file);
		gtk_window_set_title(GTK_WINDOW(window), title);
		g_free(title);
	}

	text[1] = buf;
	gtk_hlist_freeze(GTK_HLIST(icon_list));
	nres = macres_file_num_res_of_type(cicns, TYPE_cicn);
	for(icon = 0; icon < nres; icon++) {
		macres_res *r = 0;
		r = macres_file_get_nth_res_of_type(cicns, TYPE_cicn, icon);
		if(!r) continue;
		
		image = (GdkImage *)cicn_to_gdkimage(colormap, visual, r->data, 
											 r->datalen, &maskim);
		
		off = image->width > 400 ? 198 : 0;
		pixmap = gdk_pixmap_new(icon_list->window, image->width-off, 
								image->height, image->depth);
		if (!gc) gc = gdk_gc_new (pixmap);
		gdk_draw_image(pixmap, gc, image, off, 0, 0, 0, image->width-off, 
					   image->height);
		if (maskim) {
			mask = gdk_pixmap_new(icon_list->window, image->width-off, 
								  image->height, 1);
			if (!mask_gc) mask_gc = gdk_gc_new (mask);
			gdk_draw_image(mask, mask_gc, maskim, off, 0, 0, 0, 
						   image->width-off, image->height);
			gdk_image_destroy(maskim);
		}
		else
			mask = 0;
		
		gdk_image_destroy(image);
		if(!pixmap) continue;

		nfound++;
		sprintf(buf, "%u", r->resid);
		row = gtk_hlist_append(GTK_HLIST(icon_list), text);
		gtk_hlist_set_row_data_full(GTK_HLIST(icon_list), row, r->name, 
									g_free);
		gtk_hlist_set_pixtext(GTK_HLIST(icon_list), row, 0, "", 34, pixmap, 
							  mask);
		gdk_pixmap_unref(pixmap);
		if(mask) gdk_pixmap_unref(mask);
		
		g_free (r);
		
		/* Cooperative multitasking */
		if (icon && icon % 10 == 0) {
			while (gtk_events_pending()) gtk_main_iteration();
		}
	}
	if(nfound >= 2) {
		gtk_hlist_sort(GTK_HLIST(icon_list));
	}
	gtk_hlist_thaw(GTK_HLIST(icon_list));
}

static void icon_row_selected(GtkWidget *list, gint row, gint col, 
					   GdkEventButton *event, gpointer data)
{
	char *name = gtk_hlist_get_row_data(GTK_HLIST(list), row);
	char buf[256];
	
	buf[0] = '\0';
	if (name)
		snprintf(buf, sizeof(buf), "CICN Name: %s", name);
	gtk_label_set_text(GTK_LABEL(icon_label), buf);
}

static void create_icon_viewer(void)
{
	GtkWidget *scroll;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkItemFactory *item_factory;
	GtkAccelGroup *accel_group;
	gchar *titles[] = {"CICN", "CICN Id"};

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), PACKAGE);
	gtk_window_set_policy(GTK_WINDOW(window), 1, 1, 1);
	gtk_widget_set_usize(window, 330, 0);
	accel_group = gtk_accel_group_new();
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<blah>", 
										accel_group);
	gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, 0);
	gtk_accel_group_attach(accel_group, GTK_OBJECT(window));
	vbox = gtk_vbox_new(0,0);

	icon_list = gtk_hlist_new_with_titles(2, titles);
	gtk_hlist_set_selection_mode(GTK_HLIST(icon_list), GTK_SELECTION_EXTENDED);
	gtk_hlist_set_compare_func(GTK_HLIST(icon_list), 
							   (GtkHListCompareFunc) (listsorthelper));

	scroll = gtk_scrolled_window_new(0,0);
	gtk_hlist_set_column_width(GTK_HLIST(icon_list), 0, 260);
	gtk_hlist_set_column_width(GTK_HLIST(icon_list), 1, 42);
	gtk_hlist_set_row_height(GTK_HLIST(icon_list), 18);
	gtk_signal_connect(GTK_OBJECT(icon_list), "select_row", 
					   GTK_SIGNAL_FUNC(icon_row_selected), 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(scroll, 232, 256);

	gtk_container_add(GTK_CONTAINER(scroll), icon_list);

	gtk_container_add(GTK_CONTAINER(window), vbox);

	gtk_box_pack_start(GTK_BOX(vbox), 
					   gtk_item_factory_get_widget(item_factory, "<blah>"),
					   0, 0, 0);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, 0, 0, 0);

	hbox = gtk_hbox_new(0, 0);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
	icon_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), icon_label, 1, 1, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, 1, 1, 0);
	gtk_misc_set_alignment(GTK_MISC(icon_label), 0, .5);

	gtk_signal_connect(GTK_OBJECT(window), "delete_event", 
					   GTK_SIGNAL_FUNC(quit_gcicnv), 0);

	gtk_widget_show_all(window);
	
	if (cicns)
		list_icons();
}

int main (int argc, char **argv)
{
	if(argc >= 2) {
		create_ifn(argv[1]);
	}

	gtk_init(&argc, &argv);
	create_icon_viewer();
	gtk_main();

	return 0;
}
